==> oobium_datastore_gen-1.0.0.tar.gz: lib/src/generators/models_feature_model.dart <==
import 'package:collection/src/iterable_extensions.dart';
import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/generators/rest_feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/model_field.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class ModelGenerator {

  final Model m;
  Schema get schema => m.schema;
  ModelGenerator(this.m);

  String compile() => [
    compileKey,
    compileModel,
    compileRef,
    if(m.isNotSingleton) compileQuery,
  ].join('\n');

  String get compileKey =>
m.isSingleton ?
'''
class ${m.keyName} extends StateKey<${m.name}> {
  @override
  final String id = '${m.name}';
  const ${m.keyName}();
}
''' :
m.isInternal ?
'''
class ${m.keyName} extends StateKey<${m.name}> {
  @override
  final ObjectId id;
  ${m.keyName}([ObjectId? id]) : id = id ?? ObjectId();
  ${m.keyName}.parse(String s) : id = ${m.keyName}.value(s);
  static ${m.idType} value(String s) => ObjectId.fromHexString(s);
  static ${m.idType}? maybe(String? s) {
    try {
      return (s != null) ? ${m.keyName}.value(s) : null;
    } catch(_) {
      return null;
    }
  }
}
''' :
'''
class ${m.keyName} extends StateKey<${m.name}> {
  @override
  final ${m.idType} id;
  const ${m.keyName}(this.id);
  ${m.keyName}.parse(String s) : id = ${m.keyName}.value(s);
  static ${m.idType} value(String s) => ${(m.idType == 'String') ? 's' : '${m.idType}.parse(s)'};
  static ${m.idType}? maybe(String? s) {
    try {
      return (s != null) ? ${m.keyName}.value(s) : null;
    } catch(_) {
      return null;
    }
  }
}
''';

  String get compileModel =>
'''
class ${m.name} extends $modelSuper implements ${m.keyName} {
  
  ${m.name}($modelParams) : this._($modelConstructor);
  ${m.name}.fromJson(Map<String, dynamic> data, {$fromJsonParams}) : this._($fromJsonConstructor);
  static List<${m.name}> fromJsonList(list, {$fromJsonListParams}) => (list as Iterable).map((data) => ${m.name}.fromJson($fromJsonListConstructor)).toList();
  ${m.name}._($model_Constructor)$model_Super
  
  ${m.dataFields.map((f) => f.fieldDeclaration).join()}
  
  @override
  int compareTo(${m.name} other) => $compareToTests;
  
  ${m.isNotSingleton ? '@override' : ''}
  ${m.name} copyWith({$copyWithParams}) => ${m.name}._($copyWithConstructor);
  
  $toJson
}
''';

  String get compileRef =>
      m.isSingleton ? singletonRef :
      m.isExternal ? externalRef : internalRef;

  String get externalRef =>
'''
class $declaration {
  $constructor
  $accessors
  $fetch
  $update
}
''';

  String get internalRef =>
'''
class $declaration {
  $constructor
  $accessors
  $update
}
''';

  String get singletonRef =>
'''
class $declaration {
  $constructor
  $accessors
  $fetch
  $update
}
''';

  String get compileQuery =>
      '''
class ${m.queryName} {
  
  $queryFields
  
  const ${m.queryName}({$queryParams});
  ${m.queryName}._($queryParams);
  
  bool get isAll => $queryAll;
  bool get isNotAll => !isAll;
  
  bool test(${m.name} m) => $queryTest;

  @override
  bool operator ==(Object? other) => identical(this, other)
    || (other is ${m.name} && $equalsTests);
  
  @override
  int get hashCode => $objectHash;
}
''';

  String get queryFields => m.realFields.map((f) => [
    'final ${f.typeModel?.idType ?? f.type}? ${f.name};',
    if(f.isNullable) 'final bool? ${f.name}IsNull;'
  ]).expand((e) => e).join();

  String get queryParams => m.realFields.map((f) => [
    if(f.isNullable) 'this.${f.name}IsNull',
    'this.${f.name}',
  ]).expand((e) => e).join(',');

  String get queryAll => m.realFields.map((f) => [
    if(f.isNullable) '(${f.name}IsNull == null)',
    '(${f.name} == null)',
  ]).expand((e) => e).join('&&');

  String get queryTest => m.realFields.map((f) => [
    if(f.isNullable) '('
                      '(${f.name}IsNull == true) ? (m.${f.name} == null) :'
                      '(${f.name}IsNull == false) ?'
                        '(${f.name} == null) ? (m.${f.name} != null) : (${f.name} == m.${f.name}) :'
                      '((${f.name} == null) || (${f.name} == m.${f.name}))'
                     ')'
    else '(${f.name} == null || ${f.name} == m.${f.name})',
  ]).expand((e) => e).join('&&');

  String get equalsTests => m.realFields.map((f) => '${f.name} == other.${f.name}').join('&&');

  String get objectHash => m.fields.length < 20
      ? 'Object.hash($hashValues)' : 'Object.hashAll([$hashValues])';

  String get hashValues => [
    m.name,
    ...m.realFields.map((f) => f.name)
  ].join(',');

  String get modelSuper => m.isSingleton
    ? 'Singleton<${m.name}>'
    : 'Model<${m.name}, ${m.idType}>';

  String get model_Constructor => [
    if(m.isNotSingleton) '${m.idType} id',
    ...m.dataFields.map((f) => 'this.${f.name}'),
    '[ObjectId? _updateId]'
  ].join(',');

  String get model_Super => ' : super(${[
    if(m.isNotSingleton) 'id',
    '_updateId',
  ].join(',')});';

  String get modelParams => '{'
    '${m.isSingleton ? '' : m.isInternal ? 'ObjectId? id,' : 'required ${m.idType} id,'}'
    '${m.dataFields.map((f) => '${f.isRequired ? 'required ${f.jsonType}' : '${f.jsonType}${f.nullable}'} ${f.name}${f.isInitialized ? '=${f.initializer}' : ''}').join(',')}'
  '}';

  String get modelConstructor => [
    if(m.isExternal) 'id'
    else if(m.isNotSingleton) 'id ?? ObjectId()',
    ...m.dataFields.map((f) => f.name)
  ].join(',');

  String get declaration => m.isSingleton
      ? '${m.refName} extends SingletonRef<${m.name}, ${m.refName}> implements ${m.keyName}'
      : '${m.refName} extends ModelRef<${m.name}, ${m.idType}, ${m.refName}> implements ${m.keyName}';

  String get constructor =>
m.isSingleton ?
'''
  final _ModelFeature _feature;
  ${m.refName}._(this._feature) : super(_feature);
''' :
'''
  final _ModelFeature _feature;
  ${m.refName}._(this._feature, ${m.idType} id) : super(_feature, id);
''';

  String get accessors => m.fields.where((f) => f.isNotId)
      .map((f) => f.accessors).expand((e) => e).join();

  String get fetch => schema.has<RestFeature>() ?
'''

  void expire() => _feature<_RestFeature>().${m.adapterVarName}.expire($expireArgs);

  Future<${m.refName}> fetch({bool force=false}) => _feature<_RestFeature>().${m.adapterVarName}.fetch($fetchArgs).then((_) => this);
''' : '';

  String get expireArgs => m.isNotSingleton ? '${m.queryName}(id: id)' : '';

  String get fetchArgs => [
    if(m.isNotSingleton) '${m.queryName}(id: id)',
    'force'
  ].join(',');

  String get update =>
'''
  void update(${m.updateParams}) {
    final s = ${m.name}._(${m.updateConstructor});
    _feature.put(s);
  }
''';

  String get compareToTests => [
    ...m.dataFields.map((f) => 'compare(${f.name}, other.${f.name})'),
    '0',
  ].join('??');

  String get copyWithParams => [
    if(m.isNotSingleton) '${m.idType}? id',
    ...m.dataFields.map((f) => [
         '${f.jsonType}? ${f.name}',
         if(f.isNullable) 'bool ${f.name}AsNull=false'
       ]).expand((e) => e)
    ].join(',');

  String get copyWithConstructor => [
    if(m.isNotSingleton) 'id ?? this.id',
    ...m.dataFields.map((f) => f.isNullable
        ? '${f.name}AsNull ? null : (${f.name} ?? this.${f.name})'
        : '${f.name} ?? this.${f.name}')
  ].join(',');

  String get fromJsonParams => [
    if(m.isNotSingleton) '${m.idType}? id',
    ...m.dataFields.map((f) => [
      '${f.jsonType}? ${f.name}',
      if(f.isNullable) 'bool ${f.name}AsNull=false'
    ]).expand((e) => e),
    if(m.isNotSingleton) '${m.queryName}? withQuery',
  ].join(',');

  String get fromJsonConstructor => [
    if(m.isExternal) "id ?? (data['id'] as ${m.idType})"
    else if(m.isInternal) "id ?? ((data['id'] != null) ? ${m.keyName}.value(data['id']) : ObjectId())",
    ...m.dataFields.map((f) => (f.isNullable || f.isInitialized)
        ? "${fromJsonOverrides(f)} ?? ((data['${f.remote}'] != null) ? ${f.jsonDecoder} : ${f.initializer})"
        : "${fromJsonOverrides(f)} ?? ${f.jsonDecoder}")
  ].join(',');

  String fromJsonOverrides(ModelField f) => [
    f.name,
    if(m.isNotSingleton) 'withQuery?.${f.name}'
  ].join('??');

  String get fromJsonListParams => [
    if(m.isNotSingleton) '${m.idType} Function(Map<String, dynamic> item)? id',
    ...m.dataFields.map((f) => [
      '${f.jsonType}? ${f.name}',
      if(f.isNullable) 'bool ${f.name}AsNull=false'
    ]).expand((e) => e),
    if(m.isNotSingleton) '${m.queryName}? withQuery',
  ].join(',');

  String get fromJsonListConstructor => [
    'data',
    if(m.isNotSingleton) 'id: id?.call(data)',
    ...m.dataFields.map((f) => [
      '${f.name}: ${f.name}',
      if(f.isNullable) '${f.name}AsNull: ${f.name}AsNull'
    ]).expand((e) => e),
    if(m.isNotSingleton) 'withQuery: withQuery',
  ].join(',');

  String get toJson =>
  '@override '
  'Map<String, dynamic> toJson() => {${[
    if(m.isInternal) "'id': id.hexString"
    else "'id': id",
    ...m.dataFields.map((f) => f.isNullable
        ? "if(${f.name} != null) '${f.remote}': ${f.jsonEncoder}"
        : "'${f.remote}': ${f.jsonEncoder}")
    ].join(',')}'
  '};';
}

extension ModelGeneratorX on Model {

  String get updateParams => dataFields.isEmpty ? '' :
    '{${dataFields.map((f) => f.updateParameters).expand((e) => e).join(',\n')}}';

  String get updateConstructor => [
    if(isNotSingleton) 'id',
    ...dataFields.map((f) => f.isNullable
        ? f.isModel
          ? '${f.name}AsNull ? null : (${f.name} ?? this.${f.name})?.id'
          : '${f.name}AsNull ? null : (${f.name} ?? this.${f.name})'
        : f.isModel
          ? '(${f.name} ?? this.${f.name}).id'
          : '(${f.name} ?? this.${f.name})')
  ].join(',');
}

extension ModelFieldX on ModelField {

  List<String> get accessors => [
    getter,
    if(isNotHasMany) setter
  ];

  String get getter {
    if(isHasMany) {
      final type = '${model.name}${linkedModel?.namePlural}';
      return '$type get $name => $type._(_feature, this);';
    }
    final v = '_feature.get(this).$name';
    final m = typeModel;
    if(m != null) {
      if(m.isSingleton) {
        return isNullable
          ? '$refType? get $name => ($v != null) ? $refType._(_feature) : null;'
          : '$refType get $name => $refType._(_feature);';
      }
      return isNullable
        ? '$refType? get $name {'
          '  final v = $v;'
          '  return (v != null) ? $refType._(_feature, v) : null;'
          '}'
        : '$refType get $name => $refType._(_feature, $v);';
    }
    final override = isOverride ? '@override ' : '';
    return '$override$type$nullable get $name => $v;';
  }

  String get setter => isNullable
      ? 'set $name($refType? value) => (value != null)'
        '  ? update($name: value)'
        '  : update(${name}AsNull: true);'
      : 'set $name($refType value) => update($name: value);';

  String get fieldDeclaration {
    if(isHasMany) {
      throw Exception('tried to call modelDeclaration on hasMany field ($this)');
    } else {
      final type = typeModel?.idType ?? this.type;
      return 'final $type$nullable $name;';
    }
  }

  List<String> get updateParameters => [
    '$keyType? $name',
    if(isNullable) 'bool ${name}AsNull=false'
  ];

  String get jsonType => typeModel?.idType ?? type;

  String get jsonDecoder => _jsonDecoder();
  String _jsonDecoder([String? v]) {
    v ??= "data['$remote']";
    return model.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonDecoder?.replaceAll('\$', v)
        ?? model.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.decoder?.replaceAll('\$', v)
        ?? (isClass ? '$type.fromJson($v)' : null)
        ?? ((isModel && typeModel!.isInternal) ? '${typeModel!.keyName}.value($v)' : null)
        ?? (isModel ? v : null)
        ?? (isHasMany ? (throw Exception('jsonDecoder on isMany field')) : null)
        ?? (isDateTime ? 'DateTime.parse($v)' : null)
        ?? (jsonDecode?.replaceAll('\$', v))
        ?? (isCollection ? fromJsonCollection(v) : null)
        ?? (isEnum ? '$type.values.byName($v)' : null)
        ?? (isInPart ? '$type.fromJson($v)' : null)
        ?? v;
  }

  String get jsonEncoder {
    final v = local;
    final e = model.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonEncoder?.replaceAll('\$', v)
        ?? model.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.encoder?.replaceAll('\$', v)
        ?? (isHasMany ? (throw Exception('jsonEncoder on isMany field')) : null)
        ?? (jsonEncode?.replaceAll('\$', v))
        ?? (isClass ? '$v$nullable.toJson()' : null)
        ?? ((isModel && typeModel!.isInternal) ? '$v$nullable.hexString' : null)
        ?? (isModel ? v : null)
        ?? (isObjectId ? '$v$nullable.hexString' : null)
        ?? (isDateTime ? '$v$nullable.millisecondsSinceEpoch' : null)
        ?? (isPrimitive ? v : null)
        ?? (isInPart ? '$v$nullable.toJson()' : null)
        ?? (isEnum ? '$v$nullable.name' : null)
        ?? 'jsonValue($v)';
    return isNullable && isInitialized ? '$e ?? $initializer' : e;
  }

  String fromJsonCollection(String value) {
    if(isNotParameterized) {
      if(isIterable) return 'List.from($value ?? [])';
      if(isList) return 'List.from($value ?? [])';
      if(isSet) return 'Set.from($value ?? {})';
      if(isMap) return 'Map.from($value ?? {})';
      throw '$this is not a collection';
    }
    final field = ModelField.paramOf(this);
    if(field.isPrimitive) {
      if(isIterable) return 'List<${parameter}>.from($value ?? [])';
      if(isList) return 'List<${parameter}>.from($value ?? [])';
      if(isSet) return 'Set<${parameter}>.from($value ?? {})';
      if(isMap) return 'Map<${parameter}>.from($value ?? {})';
      throw '$this is not a collection';
    }
    final fromJson = field._jsonDecoder('v');
    if(isIterable) return '($value ?? []).map((v) => $fromJson).cast<$parameter>()';
    if(isList) return '($value ?? []).map((v) => $fromJson).toList().cast<$parameter>()';
    if(isSet) return '($value ?? {}).map((v) => $fromJson).toSet().cast<$parameter>()';
    if(isMap) return '($value ?? {}).map((k,v) => MapEntry(\'\$k\', $fromJson)).cast<$parameter>()';
    throw '$this is not a collection';
  }
}

==> oobium_datastore_gen-1.0.1.tar.gz: lib/src/generators/models_feature_model.dart <==
import 'package:collection/src/iterable_extensions.dart';
import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/generators/rest_feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/model_field.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class ModelGenerator {

  final Model m;
  Schema get schema => m.schema;
  ModelGenerator(this.m);

  String compile() => [
    compileKey,
    compileModel,
    compileRef,
    if(m.isNotSingleton) compileQuery,
  ].join('\n');

  String get compileKey =>
m.isSingleton ?
'''
class ${m.keyName} extends StateKey<${m.name}> {
  @override
  final String id = '${m.name}';
  const ${m.keyName}();
}
''' :
m.isInternal ?
'''
class ${m.keyName} extends StateKey<${m.name}> {
  @override
  final ObjectId id;
  ${m.keyName}([ObjectId? id]) : id = id ?? ObjectId();
  ${m.keyName}.parse(String s) : id = ${m.keyName}.value(s);
  static ${m.idType} value(String s) => ObjectId.fromHexString(s);
  static ${m.idType}? maybe(String? s) {
    try {
      return (s != null) ? ${m.keyName}.value(s) : null;
    } catch(_) {
      return null;
    }
  }
}
''' :
'''
class ${m.keyName} extends StateKey<${m.name}> {
  @override
  final ${m.idType} id;
  const ${m.keyName}(this.id);
  ${m.keyName}.parse(String s) : id = ${m.keyName}.value(s);
  static ${m.idType} value(String s) => ${(m.idType == 'String') ? 's' : '${m.idType}.parse(s)'};
  static ${m.idType}? maybe(String? s) {
    try {
      return (s != null) ? ${m.keyName}.value(s) : null;
    } catch(_) {
      return null;
    }
  }
}
''';

  String get compileModel =>
'''
class ${m.name} extends $modelSuper$mixins$interfaces {
  
  ${m.name}($modelParams) : this._($modelConstructor);
  ${m.name}.fromJson(Map<String, dynamic> data, {$fromJsonParams}) : this._($fromJsonConstructor);
  static List<${m.name}> fromJsonList(list, {$fromJsonListParams}) => (list as Iterable).map((data) => ${m.name}.fromJson($fromJsonListConstructor)).toList();
  ${m.name}._($model_Constructor)$model_Super
  
  ${m.dataFields.map((f) => f.fieldDeclaration).join()}
  
  @override
  int compareTo(${m.name} other) => $compareToTests;
  
  ${m.isNotSingleton ? '@override' : ''}
  ${m.name} copyWith({$copyWithParams}) => ${m.name}._($copyWithConstructor);
  
  $toJson
}
''';

  String get compileRef =>
      m.isSingleton ? singletonRef :
      m.isExternal ? externalRef : internalRef;

  String get externalRef =>
'''
class $declaration {
  $constructor
  $accessors
  $fetch
  $update
}
''';

  String get internalRef =>
'''
class $declaration {
  $constructor
  $accessors
  $update
}
''';

  String get singletonRef =>
'''
class $declaration {
  $constructor
  $accessors
  $fetch
  $update
}
''';

  String get compileQuery =>
      '''
class ${m.queryName} {
  
  $queryFields
  
  const ${m.queryName}({$queryParams});
  ${m.queryName}._($queryParams);
  
  bool get isAll => $queryAll;
  bool get isNotAll => !isAll;
  
  bool test(${m.name} m) => $queryTest;

  @override
  bool operator ==(Object? other) => identical(this, other)
    || (other is ${m.name} && $equalsTests);
  
  @override
  int get hashCode => $objectHash;
}
''';

  String get queryFields => m.realFields.map((f) => [
    'final ${f.typeModel?.idType ?? f.type}? ${f.name};',
    if(f.isNullable) 'final bool? ${f.name}IsNull;'
  ]).expand((e) => e).join();

  String get queryParams => m.realFields.map((f) => [
    if(f.isNullable) 'this.${f.name}IsNull',
    'this.${f.name}',
  ]).expand((e) => e).join(',');

  String get queryAll => m.realFields.map((f) => [
    if(f.isNullable) '(${f.name}IsNull == null)',
    '(${f.name} == null)',
  ]).expand((e) => e).join('&&');

  String get queryTest => m.realFields.map((f) => [
    if(f.isNullable) '('
                      '(${f.name}IsNull == true) ? (m.${f.name} == null) :'
                      '(${f.name}IsNull == false) ?'
                        '(${f.name} == null) ? (m.${f.name} != null) : (${f.name} == m.${f.name}) :'
                      '((${f.name} == null) || (${f.name} == m.${f.name}))'
                     ')'
    else '(${f.name} == null || ${f.name} == m.${f.name})',
  ]).expand((e) => e).join('&&');

  String get equalsTests => m.realFields.map((f) => '${f.name} == other.${f.name}').join('&&');

  String get objectHash => m.fields.length < 20
      ? 'Object.hash($hashValues)' : 'Object.hashAll([$hashValues])';

  String get hashValues => [
    m.name,
    ...m.realFields.map((f) => f.name)
  ].join(',');

  String get modelSuper => m.isSingleton
    ? 'Singleton<${m.name}>'
    : 'Model<${m.name}, ${m.idType}>';

  String get model_Constructor => [
    if(m.isNotSingleton) '${m.idType} id',
    ...m.dataFields.map((f) => 'this.${f.name}'),
    '[ObjectId? _updateId]'
  ].join(',');

  String get model_Super => ' : super(${[
    if(m.isNotSingleton) 'id',
    '_updateId',
  ].join(',')});';

  String get modelParams => '{'
    '${m.isSingleton ? '' : m.isInternal ? 'ObjectId? id,' : 'required ${m.idType} id,'}'
    '${m.dataFields.map((f) => '${f.isRequired ? 'required ${f.jsonType}' : '${f.jsonType}${f.nullable}'} ${f.name}${f.isInitialized ? '=${f.initializer}' : ''}').join(',')}'
  '}';

  String get modelConstructor => [
    if(m.isExternal) 'id'
    else if(m.isNotSingleton) 'id ?? ObjectId()',
    ...m.dataFields.map((f) => f.name)
  ].join(',');

  String get declaration => m.isSingleton
      ? '${m.refName} extends SingletonRef<${m.name}, ${m.refName}>$mixins$interfaces'
      : '${m.refName} extends ModelRef<${m.name}, ${m.idType}, ${m.refName}>$mixins$interfaces';

  String get mixins => m.mixins.isEmpty ? '' : ' with ${m.mixins.join(',')}';
  String get interfaces => ' implements ${[m.keyName, ...m.interfaces].join(',')}';

  String get constructor =>
m.isSingleton ?
'''
  final _ModelFeature _feature;
  ${m.refName}._(this._feature) : super(_feature);
''' :
'''
  final _ModelFeature _feature;
  ${m.refName}._(this._feature, ${m.idType} id) : super(_feature, id);
''';

  String get accessors => m.fields.where((f) => f.isNotId)
      .map((f) => f.accessors).expand((e) => e).join();

  String get fetch => schema.has<RestFeature>() ?
'''

  void expire() => _feature<_RestFeature>().${m.adapterVarName}.expire($expireArgs);

  Future<${m.refName}> fetch({bool force=false}) => _feature<_RestFeature>().${m.adapterVarName}.fetch($fetchArgs).then((_) => this);
''' : '';

  String get expireArgs => m.isNotSingleton ? '${m.queryName}(id: id)' : '';

  String get fetchArgs => [
    if(m.isNotSingleton) '${m.queryName}(id: id)',
    'force'
  ].join(',');

  String get update =>
'''
  void update(${m.updateParams}) {
    final s = ${m.name}._(${m.updateConstructor});
    _feature.put(s);
  }
''';

  String get compareToTests => [
    ...m.dataFields.map((f) => 'compare(${f.name}, other.${f.name})'),
    '0',
  ].join('??');

  String get copyWithParams => [
    if(m.isNotSingleton) '${m.idType}? id',
    ...m.dataFields.map((f) => [
         '${f.jsonType}? ${f.name}',
         if(f.isNullable) 'bool ${f.name}AsNull=false'
       ]).expand((e) => e)
    ].join(',');

  String get copyWithConstructor => [
    if(m.isNotSingleton) 'id ?? this.id',
    ...m.dataFields.map((f) => f.isNullable
        ? '${f.name}AsNull ? null : (${f.name} ?? this.${f.name})'
        : '${f.name} ?? this.${f.name}')
  ].join(',');

  String get fromJsonParams => [
    if(m.isNotSingleton) '${m.idType}? id',
    ...m.dataFields.map((f) => [
      '${f.jsonType}? ${f.name}',
      if(f.isNullable) 'bool ${f.name}AsNull=false'
    ]).expand((e) => e),
    if(m.isNotSingleton) '${m.queryName}? withQuery',
  ].join(',');

  String get fromJsonConstructor => [
    if(m.isExternal) "id ?? (data['id'] as ${m.idType})"
    else if(m.isInternal) "id ?? ((data['id'] != null) ? ${m.keyName}.value(data['id']) : ObjectId())",
    ...m.dataFields.map((f) => (f.isNullable || f.isInitialized)
        ? "${fromJsonOverrides(f)} ?? ((data['${f.remote}'] != null) ? ${f.jsonDecoder} : ${f.initializer})"
        : "${fromJsonOverrides(f)} ?? ${f.jsonDecoder}")
  ].join(',');

  String fromJsonOverrides(ModelField f) => [
    f.name,
    if(m.isNotSingleton) 'withQuery?.${f.name}'
  ].join('??');

  String get fromJsonListParams => [
    if(m.isNotSingleton) '${m.idType} Function(Map<String, dynamic> item)? id',
    ...m.dataFields.map((f) => [
      '${f.jsonType}? ${f.name}',
      if(f.isNullable) 'bool ${f.name}AsNull=false'
    ]).expand((e) => e),
    if(m.isNotSingleton) '${m.queryName}? withQuery',
  ].join(',');

  String get fromJsonListConstructor => [
    'data',
    if(m.isNotSingleton) 'id: id?.call(data)',
    ...m.dataFields.map((f) => [
      '${f.name}: ${f.name}',
      if(f.isNullable) '${f.name}AsNull: ${f.name}AsNull'
    ]).expand((e) => e),
    if(m.isNotSingleton) 'withQuery: withQuery',
  ].join(',');

  String get toJson =>
  '@override '
  'Map<String, dynamic> toJson() => {${[
    if(m.isInternal) "'id': id.hexString"
    else "'id': id",
    ...m.dataFields.map((f) => f.isNullable
        ? "if(${f.name} != null) '${f.remote}': ${f.jsonEncoder}"
        : "'${f.remote}': ${f.jsonEncoder}")
    ].join(',')}'
  '};';
}

extension ModelGeneratorX on Model {

  String get updateParams => dataFields.isEmpty ? '' :
    '{${dataFields.map((f) => f.updateParameters).expand((e) => e).join(',\n')}}';

  String get updateConstructor => [
    if(isNotSingleton) 'id',
    ...dataFields.map((f) => f.isNullable
        ? f.isModel
          ? '${f.name}AsNull ? null : (${f.name} ?? this.${f.name})?.id'
          : '${f.name}AsNull ? null : (${f.name} ?? this.${f.name})'
        : f.isModel
          ? '(${f.name} ?? this.${f.name}).id'
          : '(${f.name} ?? this.${f.name})')
  ].join(',');
}

extension ModelFieldX on ModelField {

  List<String> get accessors => [
    getter,
    if(isNotHasMany) setter
  ];

  String get getter {
    if(isHasMany) {
      final type = '${model.name}${linkedModel?.namePlural}';
      return '$type get $name => $type._(_feature, this);';
    }
    final v = '_feature.get(this).$name';
    final m = typeModel;
    if(m != null) {
      if(m.isSingleton) {
        return isNullable
          ? '$refType? get $name => ($v != null) ? $refType._(_feature) : null;'
          : '$refType get $name => $refType._(_feature);';
      }
      return isNullable
        ? '$refType? get $name {'
          '  final v = $v;'
          '  return (v != null) ? $refType._(_feature, v) : null;'
          '}'
        : '$refType get $name => $refType._(_feature, $v);';
    }
    final override = isOverride ? '@override ' : '';
    return '$override$type$nullable get $name => $v;';
  }

  String get setter => isNullable
      ? 'set $name($refType? value) => (value != null)'
        '  ? update($name: value)'
        '  : update(${name}AsNull: true);'
      : 'set $name($refType value) => update($name: value);';

  String get fieldDeclaration {
    if(isHasMany) {
      throw Exception('tried to call modelDeclaration on hasMany field ($this)');
    } else {
      final type = typeModel?.idType ?? this.type;
      return 'final $type$nullable $name;';
    }
  }

  List<String> get updateParameters => [
    '$keyType? $name',
    if(isNullable) 'bool ${name}AsNull=false'
  ];

  String get jsonType => typeModel?.idType ?? type;

  String get jsonDecoder => _jsonDecoder();
  String _jsonDecoder([String? v]) {
    v ??= "data['$remote']";
    return model.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonDecoder?.replaceAll('\$', v)
        ?? model.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.decoder?.replaceAll('\$', v)
        ?? (isClass ? '$type.fromJson($v)' : null)
        ?? ((isModel && typeModel!.isInternal) ? '${typeModel!.keyName}.value($v)' : null)
        ?? (isModel ? v : null)
        ?? (isHasMany ? (throw Exception('jsonDecoder on isMany field')) : null)
        ?? (isDateTime ? 'DateTime.parse($v)' : null)
        ?? (jsonDecode?.replaceAll('\$', v))
        ?? (isCollection ? fromJsonCollection(v) : null)
        ?? (isEnum ? '$type.values.byName($v)' : null)
        ?? (isInPart ? '$type.fromJson($v)' : null)
        ?? v;
  }

  String get jsonEncoder {
    final v = local;
    final e = model.schema.imports.firstWhereOrNull((i) => i.jsonDecodes(type))?.jsonEncoder?.replaceAll('\$', v)
        ?? model.schema.imports.firstWhereOrNull((i) => i.decodes(type))?.encoder?.replaceAll('\$', v)
        ?? (isHasMany ? (throw Exception('jsonEncoder on isMany field')) : null)
        ?? (jsonEncode?.replaceAll('\$', v))
        ?? (isClass ? '$v$nullable.toJson()' : null)
        ?? ((isModel && typeModel!.isInternal) ? '$v$nullable.hexString' : null)
        ?? (isModel ? v : null)
        ?? (isObjectId ? '$v$nullable.hexString' : null)
        ?? (isDateTime ? '$v$nullable.millisecondsSinceEpoch' : null)
        ?? (isPrimitive ? v : null)
        ?? (isInPart ? '$v$nullable.toJson()' : null)
        ?? (isEnum ? '$v$nullable.name' : null)
        ?? 'jsonValue($v)';
    return isNullable && isInitialized ? '$e ?? $initializer' : e;
  }

  String fromJsonCollection(String value) {
    if(isNotParameterized) {
      if(isIterable) return 'List.from($value ?? [])';
      if(isList) return 'List.from($value ?? [])';
      if(isSet) return 'Set.from($value ?? {})';
      if(isMap) return 'Map.from($value ?? {})';
      throw '$this is not a collection';
    }
    final field = ModelField.paramOf(this);
    if(field.isPrimitive) {
      if(isIterable) return 'List<${parameter}>.from($value ?? [])';
      if(isList) return 'List<${parameter}>.from($value ?? [])';
      if(isSet) return 'Set<${parameter}>.from($value ?? {})';
      if(isMap) return 'Map<${parameter}>.from($value ?? {})';
      throw '$this is not a collection';
    }
    final fromJson = field._jsonDecoder('v');
    if(isIterable) return '($value ?? []).map((v) => $fromJson).cast<$parameter>()';
    if(isList) return '($value ?? []).map((v) => $fromJson).toList().cast<$parameter>()';
    if(isSet) return '($value ?? {}).map((v) => $fromJson).toSet().cast<$parameter>()';
    if(isMap) return '($value ?? {}).map((k,v) => MapEntry(\'\$k\', $fromJson)).cast<$parameter>()';
    throw '$this is not a collection';
  }
}
